module m_fox
  use multiply
  include "mpif.h"

  integer :: dtype = MPI_DOUBLE_PRECISION
  integer :: err
  integer :: NORTH = 1, WEST = 2, SOUTH = 3, EAST = 4
  integer, dimension(MPI_STATUS_SIZE) :: status
  integer :: debug_rank = 0

contains
  
  subroutine fox0(A, B, C, M, nblocks, ranks, comm, rcomm, ccomm, coords)
    implicit none
    integer :: M, MM, nblocks, comm, rcomm, ccomm, step
    integer, dimension(4) :: ranks
    integer, dimension(2) :: coords
    double precision, dimension(M * M) :: A, B, C
    double precision, dimension(:), allocatable :: buf, buf2

    integer :: rank, root
    call MPI_Comm_rank(comm, rank, err)

    do step=0,nblocks - 1
       print*, 'hello'
       call multiply_add(A, B, C, M)
    enddo
    
  end subroutine fox0

  subroutine fox(A, B, C, M, nblocks, ranks, comm, rcomm, ccomm, coords)
    implicit none
    
    integer :: M, MM, nblocks, comm, rcomm, ccomm, step
    integer, dimension(4) :: ranks
    integer, dimension(2) :: coords
    double precision, dimension(M * M) :: A, B, C
    double precision, dimension(:), allocatable :: buf, buf2

    integer :: rank, root
    call MPI_Comm_rank(comm, rank, err)

    allocate(buf(M * M), buf2(M * M))

    call copy(A, buf2, M)

    MM = M * M

    do step=0,nblocks - 1
       call copy(buf2, A, M)
       root = mod(step + coords(1), nblocks)
       call MPI_Bcast(A, MM, dtype, root, rcomm, err)
       call multiply_add(A, B, C, M)
       call MPI_Sendrecv( &
            B, MM, dtype, ranks(EAST), 0, &
            buf, MM, dtype, ranks(WEST), 0, &
            comm, status, err)
       call copy(buf, B, M)
    enddo

  end subroutine fox

  subroutine copy(A, B, M)
    implicit none
    integer :: M, i
    double precision, dimension(M * M) :: A, B
    do i=1, M * M
       B(i) = A(i)
    enddo
  end subroutine copy

  subroutine generate_submatrix(A, M, offset, nblocks)
    implicit none
    
    integer :: M, i, j, iglob, jglob, nblocks
    integer, dimension(2) :: offset
    double precision, dimension(M * M) :: A
    double precision :: Mglob

    Mglob = real(M, 8) * real(nblocks, 8)
    do i=0, M - 1
       do j=0, M - 1
          iglob = offset(1) + i
          jglob = offset(2) + j
          A(1 + i * M + j) = cos(iglob * Mglob + jglob)
       enddo
    enddo
    
  end subroutine generate_submatrix

  subroutine collect_and_print(C, M, comm, nblocks)
    implicit none
    
    integer :: M, comm, nblocks, rank, other_rank
    double precision, dimension(M * M) :: C
    integer :: offset, global_offset, Mglob, i, j ,ib, jb
    integer, dimension(2) :: coords
    double precision, dimension(:), allocatable :: A, R
    integer :: src_offset, dst_offset

    Mglob = nblocks * M

    allocate(A(Mglob * Mglob), R(Mglob * Mglob))
    call MPI_Comm_rank(comm, rank, err)
    R = 42.0
    call MPI_Gather( &
         C, M * M, dtype, &
         R, M * M, dtype, &
         0, comm, err)

    if (rank.eq.0) then
       do ib=0, nblocks - 1
          do jb=0, nblocks - 1
             coords(1) = ib
             coords(2) = jb
             call MPI_Cart_rank(comm, coords, other_rank, err)
             src_offset = M * M * other_rank
             dst_offset = ib * M * Mglob + jb * M
             do i=0, M - 1
                do j=0, M - 1
                   A(1 + dst_offset + i * Mglob + j) = &
                        R(1 + src_offset + i * M + j)
                enddo
             enddo
          enddo
       enddo
       ! hack to avoid printing a jillion numbers just now
       if (Mglob.le.20) then
          do i=0, Mglob - 1
             print*, A(1 + i * Mglob:1 + i * Mglob + Mglob - 1)
          enddo
       elseif (Mglob.eq.100) then
          print*, 'Mglob 100', A(1 + 98 * Mglob + 99)
       else
          print*, A(1)
       endif
    endif

    deallocate(A)

  end subroutine collect_and_print

  subroutine generate_communicators(coords, nblocks, rank, ranks, comm, &
       rcomm, ccomm)
    implicit none
    
    integer, dimension(2) :: dimsizes, coords
    logical, dimension(2) :: wrap
    integer :: world = MPI_COMM_WORLD, comm, rcomm, ccomm, rank, nblocks
    integer, dimension(4) :: ranks ! ranks of each neighbour

    dimsizes(1) = nblocks
    dimsizes(2) = nblocks
    wrap(1) = .true.
    wrap(2) = .true.

    call MPI_Cart_create(world, 2, dimsizes, wrap, .false., comm, err)
    call MPI_Comm_rank(comm, rank, err)
    call MPI_Cart_shift(comm, 0, 1, ranks(EAST), ranks(WEST), err)
    call MPI_Cart_shift(comm, 1, 1, ranks(NORTH), ranks(SOUTH), err)
    call MPI_Cart_coords(comm, rank, 2, coords, err)
    call MPI_Comm_rank(comm, rank, err)
    call MPI_Comm_split(comm, coords(1), coords(2), rcomm, err)
    call MPI_Comm_split(comm, coords(2), coords(1), ccomm, err)

  end subroutine generate_communicators

end module m_fox
