program matrixmult
  use m_fox
  implicit none

  integer :: M, MM, i, j
  integer :: comm, ccomm, rcomm
  integer :: nblocks
  integer :: size, rank, iglob, jglob
  integer, dimension(2) :: coords, offset, other_offset
  logical, dimension(2) :: wrap

  integer :: rnorth, rwest, rsouth, reast
  integer, dimension(4) :: ranks
  
  double precision, dimension(:), allocatable :: A, B, C
  double precision :: t1, t2

  call MPI_Init(err)
  call MPI_Comm_size(MPI_COMM_WORLD, size, err)

  ! sqrt does only floats
  if (size.eq.1) then
     nblocks = 1
  elseif (size.eq.4) then
     nblocks = 2
  elseif (size.eq.9) then
     nblocks = 3
  elseif (size.eq.16) then
     nblocks = 4
  else
     print*, 'bad process count', size
     call MPI_Finalize(err)
     stop
  endif
  
  call generate_communicators(coords, nblocks, rank, ranks, comm, rcomm, ccomm)

  M = 1200
  MM = M**2
  
  allocate(A(MM), B(MM), C(MM))
  offset(1) = coords(1) * M
  offset(2) = coords(2) * M
  
  call generate_submatrix(A, M, offset, nblocks)
  call generate_submatrix(B, M, offset, nblocks)
  call  MPI_Barrier(comm, err)
  C = 0.0
  t1 = MPI_Wtime()
  !call multiply_add(A, B, C, M)
  call fox(A, B, C, M, nblocks, ranks, comm, rcomm, ccomm, coords)
  call MPI_Barrier(comm, err)
  t2 = MPI_Wtime()
  if (rank.eq.0) then
     print*, nblocks**2, t2 - t1
  endif
  call collect_and_print(C, M, comm, nblocks)

  deallocate(A, B, C)

  call MPI_Finalize(err)

end program matrixmult

subroutine printmatrix(A, M, coords, rank)
  implicit none
  
  integer :: M, i, rank
  integer, dimension(2) :: coords
  double precision, dimension(M * M) :: A
  print*, '---'
  do i=0, M - 1
     print*, rank, '(', coords, ')' , A(1 + M * i:1 + (M - 1) + M * i)
  enddo
  print*, '---'
end subroutine printmatrix
