program redblack_gaussseidel
  use redblack
  implicit none

  integer :: Mglob, Nglob, M, N, niter
  integer :: i, j
  integer :: world, comm, wgroup, group, wrank, rank, size
  integer :: Msize, Nsize
  integer, dimension(2) :: dimsizes
  logical, dimension(2) :: wrap
  integer, dimension(2) :: coords
  integer :: Mremainder, Nremainder
  integer :: Mglob_offset, Nglob_offset
  integer, dimension(2) :: sendbuf, recvbuf

  double precision, dimension(:, :), allocatable :: U
  double precision :: h2
  double precision :: t1, t2
  double precision :: maxerr

  character(8) :: filename = 'args.in'
  character(64) :: outfilename
  call read_args(filename, Mglob, Nglob, Msize, Nsize, niter, maxerr)

  dimsizes(1) = Msize
  dimsizes(2) = Nsize
  wrap(1) = .false.
  wrap(2) = .false.
  call MPI_Init(err)
  world = MPI_COMM_WORLD
  call MPI_Comm_rank(world, wrank, err)
  call MPI_Comm_size(world, size, err)
  call MPI_Barrier(world, err)
  call MPI_Cart_create(world, 2, dimsizes, wrap, .false., comm, err)
  call MPI_Comm_rank(comm, rank, err)
  call MPI_Cart_coords(comm, rank, 2, coords, err)

  Mremainder = mod(Mglob, Msize)
  Nremainder = mod(Nglob, Nsize)
  if ((Mremainder.ne.0).or.(Nremainder.ne.0).or.((Msize * Nsize).ne.size)) then
     print*, 'Bad things happen'
     call MPI_Abort(comm, 17)
  endif

  M = Mglob / Msize + 2
  N = Nglob / Nsize + 2

  h2 = (1.0 / M)**2
  allocate(U(M, N))

  do j=1, N
     do i=1, M
        U(i, j) = 0.0
     enddo
  enddo

  call MPI_Barrier(comm, err)

  t1 = MPI_Wtime()
  call solve(U, M, N, h2, niter, comm, maxerr)
  call MPI_Barrier(comm, err)
  t2 = MPI_Wtime()

  if (rank.eq.0) then
     print*, size, (t2 - t1) / niter
  endif
  
  !print*, outfilename, rank, coords(1), coords(2)
  if (M.lt.80) then
     write(outfilename, '(A,I2,A)') 'u.', rank, '.dat'
     open(10, file=outfilename)
     call savearray(U, M, N, 10, coords)
     close(10)
  endif
  deallocate(U)

  call MPI_Finalize(err)

end program redblack_gaussseidel

subroutine savearray(U, M, N, fileno, coords)
  implicit none
  
  integer, intent(in) :: fileno
  integer, intent(in) :: M, N
  double precision, dimension(M, N), intent(in) :: U
  integer, dimension(2) :: coords
  
  integer :: i, j
  
  do j=1, N
     do i=1, M
        ! integer vs real
        write(fileno, '(3E12.4)') coords(1) * M + real(i - 1), &
             coords(2) * M + real(j - 1), U(i, j)
     enddo
     write(fileno, '(A)')
  enddo
end subroutine savearray
